//
// *****************************************************************************
// |          G4LECS -- GEANT4 Low-Energy Compton Scattering package.          |
// |                          R. M. Kippen / LANL                              |
// |                                                                           |
// | G4LECSCompData.cc                                                         |
// |                                                                           |
// | Classes to read, store, and retrieve tabulated data for Compton           |
// | scattering, including Doppler broadening.                                 |
// |                                                                           |
// | RMK, DEC-2005. Migrate <strstream> to <sstream> for G4.8.0                |
// | RMK, DEC-2003. Change g4std headers to standard gcc for G4.6.0            |
// | RMK, MAY-2002. First version (based on G3LECS & G4CompositeEMDataSet.cc)  |
// *****************************************************************************
//

#include "G4LECSCompData.hh"
#include "Randomize.hh"
#include <fstream>
#include <sstream>


// *****************************************************************************
G4LECSCompData::G4LECSCompData(G4int minZ, G4int maxZ)
  :zMin(minZ), zMax(maxZ)
// *****************************************************************************
{
  nComponents = 0;
}

// *****************************************************************************
G4LECSCompData::G4LECSCompData(const G4String& dataFile,
					   G4int minZ, G4int maxZ)
  : zMin(minZ), zMax(maxZ)
// *****************************************************************************
{
  nComponents = 0;
  LoadData(dataFile);

}

// *****************************************************************************
G4LECSCompData::~G4LECSCompData()
// *****************************************************************************
{ 
  for (size_t i=0; i<nComponents; i++)    
    {
      delete components[i];
    }
}

// *****************************************************************************
G4int G4LECSCompData::RandomAtomicShell(G4int Z) const
// *****************************************************************************
{
  // Randomly select atomic sub-shell according to the number of electrons
  // in each sub-shell (aka the shell "occupancy").  The normalized cumulative
  // shell occupancy (computed previously, read from a data file) is used 
  // to do the actual random sampling.
  //
  G4int zIndex = Z-1;
  G4int nShells = components[zIndex]->nShell;
  G4double random = G4UniformRand();

  G4int iShell = 0;                         // Binary search
  G4int upperBound = nShells-1;
  while (iShell <= upperBound) {
      G4int midBin = (iShell + upperBound)/2;
      if ( random < components[zIndex]->shellSampler[midBin] ) 
	upperBound = midBin-1;
      else 
	iShell = midBin+1;
  }  
  if (iShell == nShells) iShell--;

  return iShell;
}

// *****************************************************************************
G4double G4LECSCompData::MaxTabulatedMomentum(G4int Z) const
// *****************************************************************************
{
  // Return the maximum tabulated bound momentum in Atomic Units for
  // this element Z.
  //
  G4int zIndex = Z-1;
  return components[zIndex]->pProfile[components[zIndex]->nPoints-1];

}

// *****************************************************************************
G4double G4LECSCompData::RandomProfileMomentum(G4int Z, G4int iShell) const
// *****************************************************************************
{
  // Randomly select bound electron momentum (in Atomic Units) for a 
  // particular electron subshell according to the normalized cumulative 
  // Compton profile distribution for this element and subshell.  Momentum
  // value is interpolated from tabulated profile data.
  //
  G4int zIndex = Z-1;

  return LogLogIntrp(G4UniformRand(),
		     components[zIndex]->shellDat[iShell].cpCDF, 
		     components[zIndex]->pProfile);
}

// *****************************************************************************
G4double G4LECSCompData::MeanIonizationEnergy(G4int Z) const
// *****************************************************************************
{
  // Return the mean ionization energy in MeV for this element.
  //
  G4int zIndex = Z-1;
  return components[zIndex]->meanPotential;
}

// *****************************************************************************
G4double G4LECSCompData::ShellIonizationEnergy(G4int Z, G4int iShell) const
// *****************************************************************************
{
  // Return the ionization energy in MeV for this element shell.
  //
  G4int zIndex = Z-1;
  return components[zIndex]->shellPotential[iShell];
}

// *****************************************************************************
void G4LECSCompData::PrintData() const
// *****************************************************************************
{
  G4cout << "The full data set has " << nComponents << " components" << G4endl;
  
  for (size_t i=0; i<nComponents; i++)
    {
      G4cout << "--- Component Z = " << i+1 << " ---" << G4endl;
      G4cout << "  Number of shells: " << components[i]->nShell << G4endl;
      G4cout << "  Mean ionization potential: " << 
	components[i]->meanPotential << "  MeV" << G4endl;
      G4cout << "  Number of profile grid points: " << 
	components[i]->nPoints << G4endl;
    }

}

// *****************************************************************************
void G4LECSCompData::LoadData(const G4String& fileName)
// *****************************************************************************
{ 

  // Loop over all files, one for each element Z in specified range
  //
  for (G4int Z=zMin; Z<=zMax; Z++) {

    // Build the complete string identifying the file with the data set
    //      

    //    char nameChar[100] = {""};
    //    std::ostrstream ost(nameChar, 100, std::ios::out);      
    std::ostringstream ost;      
    ost << fileName << Z << ".dat";      
    G4String name(ost.str());      
    char* path = getenv("G4LEDATA");
    if (!path) { 
      G4String excep = "G4LECSCompData - G4LEDATA environment variable not set";
      G4Exception(excep);
    }
    
    // Open file stream in proper directory
    //
    G4String pathString(path);
    G4String dirFile = pathString + "/" + name;
    std::ifstream file(dirFile);
    std::filebuf* lsdp = file.rdbuf();      
    if (! (lsdp->is_open()) ) {
      G4String excep = "G4LECSCompData - data file: " + dirFile + " not found";
      G4Exception(excep);
    }

    // Initialize new elemental data object
    //
    G4LECSElementData* eleDat = new G4LECSElementData();      
    G4int i, j;
    
    // Read and store elemental data
    //
    char line[256];
    file.getline(line,256);
    file.getline(line,256);
    file.getline(line,256);
    file.getline(line,256);
    file >> eleDat->nShell;
    file.getline(line,256);
    file.getline(line,256);
    eleDat->shellOccupancy.resize(eleDat->nShell);
    for (i=0; i<eleDat->nShell; i++) file >> eleDat->shellOccupancy[i];
    file.getline(line,256);
    file.getline(line,256);
    eleDat->shellSampler.resize(eleDat->nShell);
    for (i=0; i<eleDat->nShell; i++) file >> eleDat->shellSampler[i];
    file.getline(line,256);
    file.getline(line,256);
    eleDat->shellPotential.resize(eleDat->nShell);
    for (i=0; i<eleDat->nShell; i++) file >> eleDat->shellPotential[i];
    file.getline(line,256);
    file.getline(line,256);
    file >> eleDat->meanPotential;
    file.getline(line,256);
    file.getline(line,256);

    // Read original profile momentum grid and resampled grid
    //
    G4int nTMP;
    file >> nTMP;
    file.getline(line,256);
    file.getline(line,256);
    G4DataVector qCap(nTMP), cProf(nTMP);
    for (i=0; i<nTMP; i++) file >> qCap[i];
    file.getline(line,256);
    file.getline(line,256);

    // Define a fine log-spaced momentum grid for cumulative CDF
    // Momentum is in Atomic Units (to get keV = AU*511/137)
    //
    eleDat->nPoints = 201;
    eleDat->pProfile.resize(eleDat->nPoints);
    G4double Pmin = qCap[1];
    G4double Pmax = qCap[qCap.size()-1];
    G4double lStrt = log10(Pmin);
    G4double lStep = log10(Pmax/Pmin)/(eleDat->nPoints - 2);
    eleDat->pProfile[0] = 0.0;                   // Special first point
    for (i=1; i<eleDat->nPoints; i++) 
      eleDat->pProfile[i] = pow(10, lStrt+(i-1)*lStep);
    
    // Read, compute & store shellwise data for this element
    //
    eleDat->shellDat.resize(eleDat->nShell);
    for (G4int iShell=0; iShell<eleDat->nShell; iShell++) {

      // Read differential Compton profile for this shell
      for (j=0; j<nTMP; j++) file >> cProf[j];
      file.getline(line,256);      
      
      // Compute cumulative CDF (normalized [0->1] for each shell over a 
      // fine log-spaced momentum
      G4DataVector ydat(eleDat->nPoints);
      ydat[0] = 0.0;             // CDF(0.0) :== 0.0 (integral from 0-->0 = 0)
      for (i=1; i<eleDat->nPoints; i++)  {
	ydat[i] = ydat[i-1] + 
	  TableInteg(eleDat->pProfile[i-1], eleDat->pProfile[i], qCap, cProf);
      }
      for (i=0; i<eleDat->nPoints; i++)      // Normalize
	ydat[i] = ydat[i]/ydat[eleDat->nPoints - 1];
      
      // Store Compton profile CDF in data object
      eleDat->shellDat[iShell].cpCDF = ydat;

    }  // End shellwise loop
      
    components.push_back(eleDat);
    nComponents++;

    file.close();

  }  // End loop over elements

}

// *****************************************************************************
G4double G4LECSCompData::LogLogIntrp(G4double xval,
					const G4DataVector& xdat, 
					const G4DataVector& ydat) const
// *****************************************************************************
{
  //
  // Interpolate value of tabulated data set at point X = xval.
  // Uses linear interpolation in Log-Log space for all but the first data
  // point, which, if X[0] = zero, will be linearly interpolated.
  //

  G4double value = 0.;
  G4int nDat = xdat.size();

  // Locate nearest values in xdat array using binary search
  //
  G4double xtmp = xval;
  if (xtmp < xdat[0]) xtmp = xdat[0];
  if (xtmp > xdat[nDat-1]) xtmp = xdat[nDat-1];
  size_t lowerBound = 0;
  size_t upperBound = nDat - 2; 
  while (lowerBound <= upperBound) {
      size_t midBin = (lowerBound + upperBound)/2;
      if ( xtmp < xdat[midBin] ) upperBound = midBin-1;
      else lowerBound = midBin+1;
  }  
  size_t bin = upperBound;

  // Interpolate value of Y between two nearest values
  //
  if (bin == 0 && (xtmp <= 0. || xdat[0] <= 0.) ) {
    G4double x1 = xdat[bin];
    G4double x2 = xdat[bin+1];              // If first bin is X = Zero then
    G4double y1 = ydat[bin];                // use linear interpolation
    G4double y2 = ydat[bin+1];
    value = y1 + (y2-y1) * (xtmp-x1) / (x2-x1);
  } else {
    G4double x1 = log10(xdat[bin]);
    G4double x2 = log10(xdat[bin+1]);       // Otherwise use Log-Log
    G4double y1 = log10(ydat[bin]);         // interpolation
    G4double y2 = log10(ydat[bin+1]);
    value = y1 + (y2-y1) * (log10(xtmp)-x1) / (x2-x1);
    value = pow(10,value);
  }

  return value;
}

// *****************************************************************************
G4double G4LECSCompData::TableInteg(G4double xlim1, G4double xlim2,
					const G4DataVector& xdat, 
					const G4DataVector& ydat) const
// *****************************************************************************
{
  //
  // Integrate a tabulated function between limits xlim1-->xlim2 using 
  // the Newton-Cotes 5-point formula.  The tabulated function is evaluated
  // using linear interpolation in LOG-LOG space.
  //
  G4double dX = (xlim2-xlim1)/4.0;
  G4double y1 = LogLogIntrp(xlim1,      xdat, ydat);
  G4double y2 = LogLogIntrp(xlim1+  dX, xdat, ydat);
  G4double y3 = LogLogIntrp(xlim1+2*dX, xdat, ydat);
  G4double y4 = LogLogIntrp(xlim1+3*dX, xdat, ydat);
  G4double y5 = LogLogIntrp(xlim2,      xdat, ydat);

  G4double value = dX * (14.*(y1+y5) + 64.*(y2+y4) + 24.*y3) / 45. ;

  return value;
}
