//
// *****************************************************************************
// |          G4LECS -- GEANT4 Low-Energy Compton Scattering package.          |
// |                          R. M. Kippen / LANL                              |
// |                                                                           |
// | G4LECSCompton.cc                                                          |
// |                                                                           |
// | Classes to initialize and perform the Compton (incoherent) scattering     |
// | process, including shellwise Doppler broadening.  Algorithms based on     |
// | those of Namito, Ban & Hirayama, NIM A349, 489-494 (1994).                |
// |                                                                           |
// | *** Present code for unpolarized radiation only ***                       |
// |                                                                           |
// | RMK, JUN-2006. Make compatible with G4.8.1                                |
// |                   - New interface to G4CompositeEMDataSet                 |
// | RMK, DEC-2004. Make compatible with G4.7.0                                |
// |                   - New interface names for ParticleChange methods.       |
// | RMK, MAY-2003. Make compatible with G4.5.1                                |
// |                   - Remove CutsPerMaterialWarning                         |
// |                   - Added cuts per region (G4MaterialCutsCouple)          |
// | RMK, JUL-2002. Corrected bug in v1.00 that prevented Doppler broadening   |
// |                at large energies and large scatter angles.  Also,         |
// |                ionization energy is now deposited locally rather than     |
// |                carried with the Compton scattered electron.               |
// | RMK, MAY-2002. First version (based on G3LECS & G4LowEnergyCompton.cc)    |
// *****************************************************************************
//
#include "G4LECSCompton.hh"
#include "G4LECSCompData.hh"
#include "Randomize.hh"
#include "G4ParticleDefinition.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4ForceCondition.hh"
#include "G4Gamma.hh"
#include "G4Electron.hh"
#include "G4DynamicParticle.hh"
#include "G4VParticleChange.hh"
#include "G4ThreeVector.hh"
#include "G4EnergyLossTables.hh"
#include "G4VCrossSectionHandler.hh"
#include "G4CrossSectionHandler.hh"
#include "G4VEMDataSet.hh"
#include "G4CompositeEMDataSet.hh"
#include "G4VDataSetAlgorithm.hh"
#include "G4LogLogInterpolation.hh"
#include "G4VRangeTest.hh"
#include "G4RangeTest.hh"
#include "G4MaterialCutsCouple.hh"
#include "G4UnitsTable.hh"

// *****************************************************************************
G4LECSCompton::G4LECSCompton(const G4String& processName)
  : G4VDiscreteProcess(processName),
    lowEnergyLimit(250*eV),             
    highEnergyLimit(100*GeV),
    intrinsicLowEnergyLimit(10*eV),
    intrinsicHighEnergyLimit(100*GeV)
// *****************************************************************************
{
  if (lowEnergyLimit < intrinsicLowEnergyLimit || 
      highEnergyLimit > intrinsicHighEnergyLimit)
    {
      G4Exception("G4LECSCompton::G4LECSCompton - energy outside intrinsic process validity range");
    }

  crossSectionHandler = new G4CrossSectionHandler;

  G4VDataSetAlgorithm* scatterInterpolation = new G4LogLogInterpolation;
  G4String scatterFile = "comp/ce-sf-";
  scatterFunctionData = new G4CompositeEMDataSet(scatterInterpolation,1.,1.);
  scatterFunctionData->LoadData(scatterFile);

  G4String profileFile = "comp/ce-g4lecs-";
  profileData = new G4LECSCompData(profileFile);

  meanFreePathTable = 0;

  rangeTest = new G4RangeTest;

}
 
// *****************************************************************************
G4LECSCompton::~G4LECSCompton()
// *****************************************************************************
{
  delete meanFreePathTable;
  delete crossSectionHandler;
  delete scatterFunctionData;
  delete profileData;
  delete rangeTest;
}

// *****************************************************************************
void G4LECSCompton::BuildPhysicsTable(const G4ParticleDefinition&)
// *****************************************************************************
{
 
  crossSectionHandler->Clear();
  G4String crossSectionFile = "comp/ce-cs-";
  crossSectionHandler->LoadData(crossSectionFile);

  delete meanFreePathTable;
  meanFreePathTable = crossSectionHandler->BuildMeanFreePathForMaterials();

  PrintInfoDefinition();  

}

// *****************************************************************************
G4VParticleChange* G4LECSCompton::PostStepDoIt(const G4Track& aTrack, 
					       const G4Step&  aStep)
// *****************************************************************************
{
  // Simulates Compton scattering including atomic electron binding effects 
  // (aka, Doppler broadening).  Based heavily on the EGS4 algorithm of 
  // Namito et al. NIM A349, 489 (1994).  Uses tabulated Compton profile data 
  // from Biggs, Mendelsohn & Mann At. Data Nucl. Data Tables 16, 201 (1975)
  // to randomly sample shellwise bound electron momentum.
  //

  //-----------------------------------------------------------
  // Initialize dynamic particle quantities  
  //-----------------------------------------------------------
  aParticleChange.Initialize(aTrack);

  const G4DynamicParticle* incidentPhoton = aTrack.GetDynamicParticle();
  G4double photonEnergy0 = incidentPhoton->GetKineticEnergy();

  if (photonEnergy0 <= lowEnergyLimit)
    {
      aParticleChange.ProposeTrackStatus(fStopAndKill);
      aParticleChange.ProposeEnergy(0.);
      aParticleChange.ProposeLocalEnergyDeposit(photonEnergy0);
      return G4VDiscreteProcess::PostStepDoIt(aTrack,aStep);
    }

  //////////////////////////////////////////////////////////
  // startTest:
  //  photonEnergy0 = 1000.0*keV;
  //////////////////////////////////////////////////////////

  G4double e0m = photonEnergy0 / electron_mass_c2 ;
  G4ParticleMomentum photonDirection0 = incidentPhoton->GetMomentumDirection();
  G4double epsilon0 = 1. / (1. + 2. * e0m);
  G4double epsilon0Sq = epsilon0 * epsilon0;
  G4double alpha1 = -log(epsilon0);
  G4double alpha2 = 0.5 * (1. - epsilon0Sq);
  G4double wlPhoton = h_Planck*c_light/photonEnergy0;

  //-----------------------------------------------------------
  // Randomly select one element (Z) in the current material
  //-----------------------------------------------------------
  const G4MaterialCutsCouple* couple = aTrack.GetMaterialCutsCouple();
  G4int Z = crossSectionHandler->SelectRandomAtom(couple,photonEnergy0);

  //-----------------------------------------------------------
  // Butcher & Messel's algorithm as modified in 
  // G4LowEnergyCompton.  Samples the scattered photon energy
  // from dSigma/dOmega including the incoherent scattering
  // form factor; then computes the scatter angle sine & cosine
  //-----------------------------------------------------------
  G4double epsilon;
  G4double oneMepsilon;
  G4double epsilonSq;
  G4double oneCosT;
  G4double sinT2;
  G4double gReject;
  do {
    if ( alpha1/(alpha1+alpha2) > G4UniformRand()) {
      epsilon = exp(-alpha1 * G4UniformRand());  
      epsilonSq = epsilon * epsilon; 
    } else {
      epsilonSq = epsilon0Sq + (1. - epsilon0Sq) * G4UniformRand();
      epsilon = sqrt(epsilonSq);
    }      
    oneMepsilon = (1. - epsilon);
    oneCosT = oneMepsilon / ( epsilon * e0m);
    sinT2 = oneCosT * (2. - oneCosT);      
    G4double x = sqrt(oneCosT/2.) / (wlPhoton/cm);
    G4double scatteringFunction = scatterFunctionData->FindValue(x,Z-1);
    gReject = (1. - epsilon * sinT2 / (1. + epsilonSq)) * scatteringFunction;
    
  } while(gReject < G4UniformRand()*Z);

  //------------------------------------------------------------------
  // Scatter directions (unpolarized) for gamma, do electron later
  //------------------------------------------------------------------
  // Unpolarized (uniform) azimuth scatter angle
  G4double phi = twopi * G4UniformRand();
  G4double cosPhi = cos(phi);
  G4double sinPhi = sin(phi);

  // Kinematic gamma zenith scatter angle
  G4double cosTheta = 1. - oneCosT;
  G4double sinTheta = sqrt (sinT2);

  // Gamma scatter direction vector
  G4double gdirx = sinTheta * cosPhi;
  G4double gdiry = sinTheta * sinPhi;
  G4double gdirz = cosTheta;

  //-----------------------------------------------------------
  // Doppler broadening treatment: re-determine scattered
  // photon energy by sampling bound atomic electron momentum,
  // then inverting mass-energy conservation to get the photon 
  // energy.  N.B. The scatter ANGLES remain unchanged!!
  //-----------------------------------------------------------
  // Initialize doppler variables
  G4int nDop=0, nDopMax=1000;
  G4double ESG, ESG11 = epsilon * photonEnergy0;
  G4double shellIE;

  // Randomly sample binding momentum
 DopplerStart:
  {
    // Loop counter limit to avoid runaway computation - revert to 
    // free-electron values for scattered photon energy.
    nDop++;
    if (nDop > nDopMax) {
      ESG = ESG11;
      shellIE = 0.0;
      goto DopplerEnd;
    }

    // Select a random shell
    G4int iShell = profileData->RandomAtomicShell(Z);
    
    // Ionization energy in MeV for this shell
    shellIE = profileData->ShellIonizationEnergy(Z,iShell);
    G4double ESGMax = photonEnergy0 - shellIE;    

    // Randomly sample bound electron momentum (Atomic Units!)
    G4double pBound = profileData->RandomProfileMomentum(Z,iShell);
    
    // Compute corresponding scattered photon energy
    G4double F0 = (pBound / 137.0);
    G4double F1 = (F0 * F0);
    G4double F2 = (1. + oneCosT * e0m);
    G4double F3 = (F2*F2 - F1);
    G4double F4 = (F2 - F1*cosTheta);
    G4double F5 = (F4*F4 - F3 + F1*F3);    
    if (F5 < 0.0) goto DopplerStart;
    G4double F6 = sqrt(F5);        
    G4double F7 = (photonEnergy0 / F3);            
    if (G4UniformRand() < 0.5)           // Randomly select one of 
      ESG = (F4 - F6)*F7;                // two possible roots
    else
      ESG = (F4 + F6)*F7;	
    if (ESG < 0.0 || ESG > ESGMax) goto DopplerStart;

    if (ESG < G4UniformRand()*ESGMax) goto DopplerStart;
  }
 DopplerEnd:

  //-----------------------------------------------------------
  // Successful sampling of doppler broadened scattered photon 
  // energy.  Conserve E for scattered electron kinetic energy.
  // N.B. Binding energy will be deposited in the local medium 
  // and not carried by the scattered electron.
  //-----------------------------------------------------------
  G4double photonEnergy1 = ESG;
  G4double eKineticEnergy = photonEnergy0 - photonEnergy1 - shellIE;
  G4double eTotalEnergy = eKineticEnergy + electron_mass_c2;

  //-----------------------------------------------------------
  // Update G4VParticleChange for the scattered photon 
  //-----------------------------------------------------------
  G4ThreeVector photonDirection1(gdirx,gdiry,gdirz);

  photonDirection1.rotateUz(photonDirection0);

  aParticleChange.ProposeMomentumDirection(photonDirection1) ;

  if (photonEnergy1 > 0.) {
    aParticleChange.ProposeEnergy(photonEnergy1) ;
  } else {    
    aParticleChange.ProposeEnergy(0.) ;
    aParticleChange.ProposeTrackStatus(fStopAndKill);
  }
  
  //-----------------------------------------------------------
  // Kinematics of the scattered electron 
  // Generate the electron only if with large enough range 
  // w.r.t. cuts and safety; deposit binding energy locally
  //-----------------------------------------------------------
  G4double ESE = photonEnergy0 * oneMepsilon + electron_mass_c2;    
  G4double psqd = ESE*ESE - electron_mass_c2*electron_mass_c2;
  G4double cosThetaE, sinThetaE;
  if (psqd <= 0.0) {
    cosThetaE = 0.0;
    sinThetaE = -1.0;
  } else {
    cosThetaE = (eTotalEnergy+photonEnergy1)*oneMepsilon/sqrt(psqd);
    sinThetaE = -1.0 * sqrt(1. - cosThetaE*cosThetaE);
  }

  //////////////////////////////////////////////////////////
  //   G4double ThetaG = acos(cosTheta)/deg;
  //   G4double ThetaE = acos(cosThetaE)/deg;
  //   G4cout << std::setw( 9) << photonEnergy0/keV << " "
  //   	 << std::setw( 9) << ThetaG << " "
  //   	 << std::setw( 9) << ThetaE << " "
  //   	 << std::setw( 9) << photonEnergy1/keV << " " 
  //   	 << std::setw( 9) << eKineticEnergy/keV << G4endl;
  //   goto startTest;
  //////////////////////////////////////////////////////////

  G4double edirx = sinThetaE * cosPhi;
  G4double ediry = sinThetaE * sinPhi;
  G4double edirz = cosThetaE;

  G4double safety = aStep.GetPostStepPoint()->GetSafety();

  if (rangeTest->Escape(G4Electron::Electron(),couple,eKineticEnergy,safety))
    {
      G4ThreeVector eDirection(edirx,ediry,edirz);

      eDirection.rotateUz(photonDirection0);

      G4DynamicParticle* electron = 
	new G4DynamicParticle (G4Electron::Electron(),
			       eDirection,eKineticEnergy) ;
      aParticleChange.SetNumberOfSecondaries(1);
      aParticleChange.AddSecondary(electron);
      aParticleChange.ProposeLocalEnergyDeposit(shellIE); 
    }
  else
    {
      aParticleChange.SetNumberOfSecondaries(0);
      aParticleChange.ProposeLocalEnergyDeposit(eKineticEnergy+shellIE);
    }
  
  return G4VDiscreteProcess::PostStepDoIt( aTrack, aStep);
}

// *****************************************************************************
G4bool G4LECSCompton::IsApplicable(const G4ParticleDefinition& particle)
// *****************************************************************************
{
  return ( &particle == G4Gamma::Gamma() ); 
}

// *****************************************************************************
G4double G4LECSCompton::GetMeanFreePath(const G4Track& track, 
					     G4double, 
					     G4ForceCondition*)
// *****************************************************************************
{
  const G4DynamicParticle* photon = track.GetDynamicParticle();
  G4double energy = photon->GetKineticEnergy();
  const G4MaterialCutsCouple* couple = track.GetMaterialCutsCouple();
  size_t materialIndex = couple->GetIndex();

  G4double meanFreePath;
  if (energy > highEnergyLimit) meanFreePath = meanFreePathTable->FindValue(highEnergyLimit,materialIndex);
  else if (energy < lowEnergyLimit) meanFreePath = DBL_MAX;
  else meanFreePath = meanFreePathTable->FindValue(energy,materialIndex);
  return meanFreePath;
}

// *****************************************************************************
void G4LECSCompton::PrintInfoDefinition()
// *****************************************************************************
{
                     
  G4cout << G4endl << GetProcessName() << ":  "
	 << "Compton scattering with shellwise Doppler broadening.\n"
	 << "        Uses tabulated form factors and Compton profiles. \n"
         << "        PhysicsTables from "
	           << G4BestUnit(lowEnergyLimit,"Energy")
         << " to " << G4BestUnit(highEnergyLimit,"Energy") << G4endl;

}         
